#!/usr/bin/env python
# -*- coding: utf-8 -*-

import brewery.utils as utils

class NodeFinished(Exception):
    """Exception raised when node has no active outputs - each output node signalised that it
    requires no more data."""
    pass

class Node(object):
    """Base class for procesing node
    
    .. abstract_node
    """
    def __init__(self):
        """Creates a new data processing node.
        
        :Attributes:
            * `inputs`: input pipes
            * `outputs`: output pipes
            * `description`: custom node annotation
        """

        super(Node, self).__init__()
        self.inputs = []
        self.outputs = []
        self._active_outputs = []
        self.description = None

    def initialize(self):
        """Initializes the node. Initialization is separated from creation. Put any Node subclass
        initialization in this method. Default implementation does nothing."""
        pass

    def finalize(self):
        """Finalizes the node. Default implementation does nothing."""
        pass

    def run(self):
        """Main method for running the node code. Subclasses should implement this method.
        """
        
        raise NotImplementedError("Subclasses of Node should implement the run() method")
        
    @property
    def input(self):
        """Return single node imput if exists. Convenience property for nodes which process only one
        input. Raises exception if there are no inputs or are more than one imput."""
        
        if len(self.inputs) == 1:
            return self.inputs[0]
        else:
            raise Exception("Single input requested. Node has none or more than one input (%d)."
                                    % len(self.inputs))
    
    def add_input(self, pipe):
        if pipe not in self.inputs:
            self.inputs.append(pipe)
        else:
            raise Exception("Input %s already connected" % pipe)

    def add_output(self, pipe):
        if pipe not in self.outputs:
            self.outputs.append(pipe)
        else:
            raise Exception("Output %s already connected" % pipe)
    
    def put(self, obj):
        """Put row into all output pipes.
        
        Raises `NodeFinished` exception when node's target nodes are not receiving data anymore.
        In most cases this exception might be ignored, as it is handled in the node thread
        wrapper. If you want to perform necessary clean-up in the `run()` method before exiting,
        you should handle this exception and then re-reaise it or just simply return from `run()`.

        This method can be called only from node's `run()` method. Do not call it from
        `initialize()` or `finalize()`.
        """
        active_outputs = 0
        for output in self.outputs:
            if not output.closed():
                output.put(obj)
                active_outputs += 1
        
        # This is not very safe, as run() might not expect it
        if not active_outputs:
            raise NodeFinished

    def put_record(self, obj):
        """Put record into all output pipes. Convenience method. Not recommended to be used.
        
        .. warning::
        
            Depreciated.
            
        """
        for output in self.outputs:
            output.put_record(obj)

    @property
    def input_fields(self):
        """Return fields from input pipe, if there is one and only one input pipe."""
        return self.input.fields
        
    @property
    def output_fields(self):
        """Return fields passed to the output by the node.
        
        Subclasses should override this method. Default implementation returns same fields as
        input has, raises exception when there are more inputs or if there is no input
        connected."""
        if not len(self.inputs) == 1:
            raise ValueError("Can not get default list of output fields: node has more than one input"
                             " or no input is provided. Subclasses should override this method")

        if not self.input.fields:
            raise ValueError("Can not get default list of output fields: input pipe fields are not "
                             "initialized")

        return self.input.fields
    
    @property
    def output_field_names(self):
        """Convenience method for gettin names of fields generated by the node. For more information
        see :meth:`brewery.nodes.Node.output_fields`"""
        return self.output_fields.names()

    @classmethod
    def subclasses(cls, abstract = False):
        """Get all subclasses of node.
        
        :Parameters:
            * `abstract`: If set to ``True`` all abstract classes are included as well. Default is
              ``False``
        """
        classes = []
        for c in utils.subclass_iterator(cls):
            try:
                info = getattr(c, "__node_info__")
                node_type = info.get("type")
                if node_type != "abstract":
                    classes.append(c)
            except AttributeError:
                pass

        return classes

    @classmethod
    def class_dictionary(cls):
        """Return a dictionary containing node name as key and node class as value."""
        
        classes = cls.subclasses()
        dictionary = {}
        
        for c in classes:
            try:
                name = c.identifier()
                dictionary[name] = c
            except AttributeError:
                pass

        return dictionary
        
    @classmethod
    def identifier(cls):
        """Returns an identifier name of the node class. Identifier is used for construction
        of streams from dictionaries or for any other out-of-program constructions.
        
        Node identifier is specified in the `__node_info__` dictioanry as ``name``. If no explicit
        identifier is specified, then decamelized class name will be used with `node` suffix
        removed. For example: ``CSVSourceNode`` will be ``csv_source``.
        """
        
        info = getattr(cls, "__node_info__")
        ident = None
        if info:
            ident = info.get("name")
            
        if not ident:
            ident = utils.to_identifier(utils.decamelize(cls.__name__))
            if ident.endswith("_node"):
                ident = ident[:-5]
                
        return ident
            
        
        
    def configure(self, config, safe = False):
        """Configure node.
        
        :Parameters:
            * `config` - a dictionary containing node attributes as keys and values as attribute
              values. Key ``type`` is ignored as it is used for node creation.
            * `safe` - if set to ``True`` only non-protected attributes are set. Attempt
              to set protected attribute will result in an exception. Use `safe` when you are
              configuring nodes from a user interface or a custom tool. Default is ``False``: all
              attributes can be set.
              
        Note: Handling of `safe` flag is not yet implemented, that means that any attribute can be
        set.
        """
        # FIXME: do some checking for protected attributes
        for key, value in config.items():
            setattr(self, key, value)

class SourceNode(Node):
    """Abstract class for all source nodes
    
    All source nodes should provide an attribute or implement a property (``@property``) called
    ``output_fields``.
    
    .. abstract_node
    
    """
    def __init__(self):
        super(SourceNode, self).__init__()
        # self.fields = None
    @property
    def output_fields(self):
        raise NotImplementedError("SourceNode subclasses should implement output_fields")

    def add_input(self, pipe):
        raise Exception("Should not add input pipe to a source node")

class TargetNode(Node):
    """Abstract class for all target nodes
    
    .. abstract_node
    
    """
    def __init__(self):
        super(TargetNode, self).__init__()
        self.fields = None

    @property
    def output_fields(self):
        raise RuntimeError("Output fields asked from a target object.")

    def add_output(self, pipe):
        raise RuntimeError("Should not add output pipe to a target node")
    


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import csv
import codecs
import cStringIO
import base
import brewery.metadata

class UTF8Recoder(object):
    """
    Iterator that reads an encoded stream and reencodes the input to UTF-8

    From: <http://docs.python.org/lib/csv-examples.html>
    """
    def __init__(self, f, encoding=None):
        if encoding:
            self.reader = codecs.getreader(encoding)(f)
        else: # already unicode so just return f
            self.reader = f

    def __iter__(self):
        return self

    def next(self):
        return self.reader.next().encode('utf-8')

class UnicodeReader:
    """
    A CSV reader which will iterate over lines in the CSV file "f",
    which is encoded in the given encoding.
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        f = UTF8Recoder(f, encoding)
        self.reader = csv.reader(f, dialect=dialect, **kwds)

    def next(self):
        row = self.reader.next()
        return [unicode(s, "utf-8") for s in row]

    def __iter__(self):
        return self

class UnicodeWriter:
    """
    A CSV writer which will write rows to CSV file "f",
    which is encoded in the given encoding.

    From: <http://docs.python.org/lib/csv-examples.html>
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        # Redirect output to a queue
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        new_row = []
        for value in row:
            if type(value) == unicode or type(value) == str:
                new_row.append(value.encode("utf-8"))
            elif value:
                new_row.append(unicode(value))
            else:
                new_row.append(None)
                
        self.writer.writerow(new_row)
        # Fetch UTF-8 output from the queue ...
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        # ... and reencode it into the target encoding
        data = self.encoder.encode(data)
        # write to the target stream
        self.stream.write(data)
        # empty queue
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)

class CSVDataSource(base.DataSource):
    """docstring for ClassName
    
    Some code taken from OKFN Swiss library.
    """
    def __init__(self, resource, read_header = True, dialect = None, encoding=None, detect_encoding = False, 
                detect_header = False, sample_size = 200, skip_rows = None, **reader_args):
        """Creates a CSV data source stream.
        
        :Attributes:
            * resource: file name, URL or a file handle with CVS data
            * read_header: flag determining whether first line contains header or not. 
                ``True`` by default.
            * encoding: source character encoding, by default no conversion is performed. 
            * detect_encoding: read sample from source and determine whether source is UTF8 or not
            * detect_headers: try to determine whether data source has headers in first row or not
            * sample_size: maximum bytes to be read when detecting encoding and headers in file. By
                default it is set to 200 bytes to prevent loading huge CSV files at once.
            * skip_rows: number of rows to be skipped. Default: ``None``
            
        Note: avoid auto-detection when you are reading from remote URL stream.
        
        """
        self.read_header = read_header
        self.encoding = encoding
        self.detect_encoding = detect_encoding
        self.detect_header = detect_header
        
        self._autodetection = detect_encoding or detect_header
        
        self.sample_size = sample_size
        self.resource = resource
        self.reader_args = reader_args
        self.reader = None
        self.dialect = dialect
        
        self.close_file = False
        self.skip_rows = skip_rows
        self._fields = None
        
    def initialize(self):
        """Initialize CSV source stream:
        
        #. perform autodetection if required:
            #. detect encoding from a sample data (if requested)
            #. detect whether CSV has headers from a sample data (if requested)
        #.  create CSV reader object
        #.   read CSV headers if requested and initialize stream fields
        
        """

        self.file, self.close_file = base.open_resource(self.resource)

        handle = None
        
        if self._autodetection:
            
            sample = self.file.read(self.sample_size)

            # Encoding test
            if self.detect_encoding and type(sample) == unicode:
                self.encoding = "utf-8"

            if self.detect_header:
                sample = sample.encode('utf-8')
                sniffer = csv.Sniffer()
                self.read_header = sniffer.has_header(sample)

            self.file.seek(0)
            
        if self.dialect:
            if type(self.dialect) == str:
                dialect = csv.get_dialect(self.dialect)
            else:
                dialect = self.dialect
                
            self.reader_args["dialect"] = dialect

        # self.reader = csv.reader(handle, **self.reader_args)
        self.reader = UnicodeReader(self.file, encoding = self.encoding, 
                                    **self.reader_args)

        if self.skip_rows:
            for i in range(0, self.skip_rows):
                self.reader.next()
                
        # Initialize field list
        if self.read_header:
            field_names = self.reader.next()
            
            fields = [ (name, "string", "default") for name in field_names]
            
            self._fields = brewery.metadata.FieldList(fields)
        
    def finalize(self):
        if self.file and self.close_file:
            self.file.close()

    def rows(self):
        if not self.reader:
            raise RuntimeError("Stream is not initialized")
        if not self.fields:
            raise RuntimeError("Fields are not initialized")
        return self.reader

    def records(self):
        fields = self.field_names
        for row in self.reader:
            yield dict(zip(fields, row))

class CSVDataTarget(base.DataTarget):
    def __init__(self, resource, write_headers = True, truncate = True, encoding = "utf-8", 
                dialect = None, **kwds):
        """Creates a CSV data target
        
        :Attributes:
            * resource: target object - might be a filename or file-like object
            * write_headers: write field names as headers into output file
            * truncate: remove data from file before writing, default: True
            
        """
        self.resource = resource
        self.write_headers = write_headers
        self.truncate = truncate
        self._fields = None
        self.encoding = encoding
        self.dialect = dialect
        self.kwds = kwds

        self.close_file = False
        self.file = None
    def initialize(self):
        if self.truncate:
            mode = "w"
        else:
            mode = "a"

        self.file, self.close_file = base.open_resource(self.resource, mode)

        self.writer = UnicodeWriter(self.file, encoding = self.encoding, 
                                    dialect = self.dialect, **self.kwds)
        
        if self.write_headers:
            self.writer.writerow(self.field_names)

    def finalize(self):
        if self.file and self.close_file:
            self.file.close()

    def append(self, obj):
        if type(obj) == dict:
            row = []
            for field in self.field_names:
                row.append(obj.get(field))
        else:
            row = obj
                
        self.writer.writerow(row)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Brewery"""

from metadata import *

__version__ = '0.7.0'

brewery_search_paths = ['/etc/brewery', \
						'~/.brewery/', \
						'./.brewery/']

def set_brewery_search_paths(paths):
    global brewery_search_paths
    brewery_search_paths = paths

def default_logger_name():
    return 'brewery'

__all__ = (
    "Field",
    "FieldList",
    "fieldlist",
    "expand_record",
    "collapse_record",
    "default_logger_name",
    "set_brewery_search_paths",
    "__version__"
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import base
import brewery.metadata

try:
    import sqlalchemy
except:
    from brewery.utils import MissingPackage
    sqlalchemy = MissingPackage("sqlalchemy", "SQL streams", "http://www.sqlalchemy.org/",
                                comment = "Recommended version is > 0.7")
    _sql_to_brewery_types = ()
    _brewery_to_sql_type = {}

if sqlalchemy:
    # (sql type, storage type, analytical type)
    _sql_to_brewery_types = (
        (sqlalchemy.types.UnicodeText, "text", "typeless"),
        (sqlalchemy.types.Text, "text", "typeless"),
        (sqlalchemy.types.Unicode, "string", "set"),
        (sqlalchemy.types.String, "string", "set"),
        (sqlalchemy.types.Integer, "integer", "discrete"),
        (sqlalchemy.types.Numeric, "float", "range"),
        (sqlalchemy.types.DateTime, "date", "typeless"),
        (sqlalchemy.types.Date, "date", "typeless"),
        (sqlalchemy.types.Time, "unknown", "typeless"),
        (sqlalchemy.types.Interval, "unknown", "typeless"),
        (sqlalchemy.types.Boolean, "boolean", "flag"),
        (sqlalchemy.types.Binary, "unknown", "typeless")
    )

    _brewery_to_sql_type = {
        "string": sqlalchemy.types.Unicode,
        "text": sqlalchemy.types.UnicodeText,
        "date": sqlalchemy.types.Date,
        "time": sqlalchemy.types.DateTime,
        "integer": sqlalchemy.types.Integer,
        "float": sqlalchemy.types.Numeric,
        "boolean": sqlalchemy.types.SmallInteger
    }

def split_table_schema(table_name):
    """Get schema and table name from table reference.

    Returns: Tuple in form (schema, table)
    """

    split = table_name.split('.')
    if len(split) > 1:
        return (split[0], split[1])
    else:
        return (None, split[0])

class SQLDataStore(object):
    def __init__(self, url=None, connection=None, schema=None, **options):
        if connection:
            self.connection = connection
            self.engine = self.connection.engine
            self.close_connection = False
        else:
            self.engine = sqlalchemy.create_engine(url, **options)
            self.connection = self.engine.connect()
            self.close_connection = True

        self.metadata = sqlalchemy.MetaData()
        self.metadata.bind = self.engine
        self.schema = schema

    def close(self):
        if self.close_connection:
            self.connection.close()

    def dataset(self, name):
        return SQLDataset(self._table(name))

    def has_dataset(self, name):
        table = self._table(name, autoload=False)
        return table.exists()

    def create_dataset(self, name, fields, replace=False,
                       add_id_key=False, id_key_name=None):
        """Create a table."""

        if self.has_dataset(name):
            if not replace:
                raise ValueError("Dataset '%s' already exists" % name)
            else:
                table = self._table(name, autoload=False)
                table.drop(checkfirst=False)

        table = self._table(name, autoload=False)

        if add_id_key:
            if not id_key_name:
                id_key_name = 'id'

            sequence_name = "seq_" + name + "_" + id_key_name
            sequence = sqlalchemy.schema.Sequence(sequence_name, optional=True)

            col = sqlalchemy.schema.Column(id_key_name, sqlalchemy.types.Integer,
                                            sequence, primary_key=True)
            table.append_column(col)

        for field in fields:
            if not isinstance(field, brewery.metadata.Field):
                raise ValueError("field %s is not subclass of brewery.metadata.Field" % (field))

            concrete_type = field.concrete_storage_type

            if not isinstance(concrete_type, sqlalchemy.types.TypeEngine):
                concrete_type = _brewery_to_sql_type.get(field.storage_type)
                if not concrete_type:
                    raise ValueError("unable to find concrete storage type for field '%s' "
                                     "of type '%s'" % (field.name, field.storage_type))

            col = sqlalchemy.schema.Column(field.name, concrete_type)
            table.append_column(col)

        table.create()

        dataset = SQLDataset(table)
        return dataset

    def _table(self, name, autoload=True):
        split = split_table_schema(name)
        schema = split[0]
        table_name = split[1]

        if not schema:
            schema = self.schema

        table = sqlalchemy.Table(table_name, self.metadata, autoload=autoload, schema=schema)
        return table

class SQLDataset(object):
    def __init__(self, table):
        super(SQLDataset, self).__init__()
        self.table = table
        self._fields = None

    @property
    def field_names(self):
        names = [column.name for column in self.table.columns]
        return names

    @property
    def fields(self):
        if self._fields:
            return self._fields

        fields = []
        for column in self.table.columns:
            field = base.Field(name=column.name)
            field.concrete_storage_type = column.type

            for conv in _sql_to_brewery_types:
                if issubclass(column.type.__class__, conv[0]):
                    field.storage_type = conv[1]
                    field.analytical_type = conv[2]
                    break

            if not field.storage_type:
                field.storaget_tpye = "unknown"

            if not field.analytical_type:
                field.analytical_type = "unknown"

            fields.append(field)

        self._fields = fields

        return fields

class SQLDataSource(base.DataSource):
    """docstring for ClassName
    """
    def __init__(self, connection=None, url=None,
                    table=None, statement=None, schema=None, **options):
        """Creates a relational database data source stream.
        
        :Attributes:
            * url: SQLAlchemy URL - either this or connection should be specified
            * connection: SQLAlchemy database connection - either this or url should be specified
            * table: table name
            * statement: SQL statement to be used as a data source (not supported yet)
            * options: SQL alchemy connect() options
        
        Note: avoid auto-detection when you are reading from remote URL stream.
        
        """

        if not url and not connection:
            raise AttributeError("Either url or connection should be provided for SQL data source")

        if not table and not statement:
            raise AttributeError("Either table or statement should be provided for SQL data source")

        if statement:
            raise NotImplementedError("SQL source stream based on statement is not yet implemented")

        if not options:
            options = {}

        self.url = url
        self.connection = connection
        self.table_name = table
        self.statement = statement
        self.schema = schema
        self.options = options

        self._fields = None

    def initialize(self):
        """Initialize source stream:
        """
        self.datastore = SQLDataStore(self.url, self.connection, self.schema, **self.options)
        self.dataset = self.datastore.dataset(self.table_name)

    def finalize(self):
        self.datastore.close()

    @property
    def fields(self):
        if self._fields:
            return self._fields
        self._fields = self.dataset.fields
        return self._fields

    def read_fields(self):
        self._fields = self.dataset.fields
        return self._fields

    def rows(self):
        if not self.dataset:
            raise RuntimeError("Stream is not initialized")
        return self.dataset.table.select().execute()

    def records(self):
        if not self.dataset:
            raise RuntimeError("Stream is not initialized")
        fields = self.field_names
        for row in self.rows():
            record = dict(zip(fields, row))
            yield record

class SQLDataTarget(base.DataTarget):
    """docstring for ClassName
    """
    def __init__(self, connection=None, url=None,
                    table=None, schema=None, truncate=False,
                    create=False, replace=False,
                    add_id_key=False, id_key_name=None,
                    buffer_size=None, **options):
        """Creates a relational database data target stream.
        
        :Attributes:
            * url: SQLAlchemy URL - either this or connection should be specified
            * connection: SQLAlchemy database connection - either this or url should be specified
            * table: table name
            * truncate: whether truncate table or not
            * create: whether create table on initialize() or not
            * replace: Set to True if creation should replace existing table or not, otherwise
              initialization will fail on attempt to create a table which already exists.
            * options: other SQLAlchemy connect() options
            * add_id_key: whether to add auto-increment key column or not. Works only if `create`
              is ``True``
            * id_key_name: name of the auto-increment key. Default is 'id'
            * buffer_size: size of INSERT buffer - how many records are collected before they are
              inserted using multi-insert statement. Default is 1000
        
        Note: avoid auto-detection when you are reading from remote URL stream.
        
        """
        if not url and not connection:
            raise AttributeError("Either url or connection should be provided for SQL data source")

        if not options:
            options = {}

        self.url = url
        self.connection = connection
        self.table_name = table
        self.schema = schema
        self.options = options
        self.replace = replace
        self.create = create
        self.truncate = truncate
        self.add_id_key = add_id_key

        self._fields = None

        if id_key_name:
            self.id_key_name = id_key_name
        else:
            self.id_key_name = 'id'

        if buffer_size:
            self.buffer_size = buffer_size
        else:
            self.buffer_size = 1000

    def initialize(self):
        """Initialize source stream:
        """

        self.datastore = SQLDataStore(self.url, self.connection, self.schema, **self.options)

        if self.create:
            self.dataset = self.datastore.create_dataset(self.table_name,
                                                         self.fields,
                                                         self.replace,
                                                         self.add_id_key, self.id_key_name)
        else:
            self.dataset = self.datastore.dataset(self.table_name)

        if self.truncate:
            self.dataset.table.delete().execute()

        if not self._fields:
            self._fields = self.dataset.fields

        self.insert_command = self.dataset.table.insert()
        self._buffer = []

    def finalize(self):
        """Closes the stream, flushes buffered data"""

        self._flush()
        self.datastore.close()

    def __get_fields(self):
        return self._fields

    def __set_fields(self, fields):
        self._fields = fields

    fields = property(__get_fields, __set_fields)

    def __append(self, obj):
        if type(obj) == dict:
            record = obj
        else:
            record = dict(zip(self.field_names, obj))
            # if self.add_id_key:
            #     record[self.id_key_name] = None
        self.insert_command.execute(record)

    def append(self, obj):
        if type(obj) == dict:
            record = obj
        else:
            record = dict(zip(self.field_names, obj))

        self._buffer.append(record)
        if len(self._buffer) >= self.buffer_size:
            self._flush()

    def _flush(self):
        if len(self._buffer) > 0:
            self.datastore.connection.execute(self.insert_command, self._buffer)
            self._buffer = []



class MissingValuesProbe(object):
    """Data quality statistics for a dataset field
    
    :Attributes:
        * `count`: total count of null records
    """    
    def __init__(self):
        self.count = 0
        
    def probe(self, value):
        """Probe the value.
        """
        
        if value is None:
            self.count += 1

class StatisticsProbe(object):
    """Data quality statistics for a dataset field

    :Attributes:
        * `record_count`: total count of records in dataset. This should be set explicitly on
          finalisation. Seet :meth:`FieldStatistics.finalize`. In relational database this should be the
          same as `value_count`.

        * `fields`: 
        * `count`: number of non-null objects
    """    
    def __init__(self):
        self.min = None
        self.max = None
        self.sum = None
        self.count = 0
    @property
    def average(self):
        return self.sum / self.count
        
    def probe(self, value):
        self.count += 1
        if value is not None:
            if self.sum is None:
                self.sum = value
                self.min = value
                self.max = value
            else:
                self.sum += value
                self.min = min(self.min, value)
                self.max = max(self.max, value)

class DistinctProbe(object):
    def __init__(self, threshold = None):
        self.distinct_values = set([])
        self.overflow = False
        self.threshold = threshold

    def probe(self, value):
        if self.threshold and len(self.distinct_values) >= threshold:
            self.overflow = True
            return

        self.distinct_values.add(value)

class StorageTypeProbe(object):
    """Probe for guessing field data type

    Attributes:
        * field: name of a field which statistics are being presented
        * storage_types: found storage types
        * unique_storage_type: if there is only one storage type, then this is set to that type
    """    
    def __init__(self):
        self.storage_types = set()

    def probe(self, value):
        storage_type = value.__class__
        self.storage_types.add(storage_type.__name__)

    @property
    def unique_storage_type(self):
        """Return storage type if there is only one. This should always return a type in relational
        databases, but does not have to in databases such as MongoDB."""

        if len(self.storage_types) == 1:
            return list(self.storage_types)[0]
        else:
            return None
    


import copy

def fieldlist(fields):
    """Create a :class:`FieldList` from a list of strings, dictionaries or tuples.

    How fields are constructed:

    * string: `field name` is set 
    * tuple: (`field_name`, `storaget_type`, `analytical_type`), the `field_name` is
      obligatory, rest is optional
    * dict: contains key-value pairs for initializing a :class:`Field` object

    For strings and in if not explicitly specified in a tuple or a dict case, then following rules
    apply:

    * `storage_type` is set to ``unknown``
    * `analytical_type` is set to ``typeless``
    """

    return FieldList(fields)
    
# FIXME: Depreciated
def field_name(field):
    """Return a field name. If the `field` is a string object, return just the string. If 
    the `field` is `Field` instance then return `field.name` 
    
    Warning: Depreciated.
    
    """
    if type(field) == str or type(field) == unicode:
        return field
    else:
        return field.name
    
def expand_record(record, separator = '.'):
    """Expand record represented as dict object by treating keys as key paths separated by
    `separator`, which is by default ``.``. For example: ``{ "product.code": 10 }`` will become
    ``{ "product" = { "code": 10 } }``
    
    See :func:`brewery.collapse_record` for reverse operation.
    """
    result = {}
    for key, value in record.items():
        current = result
        path = key.split(separator)
        for part in path[:-1]:
            if part not in current:
                current[part] = {}
            current = current[part]
        current[path[-1]] = value
    return result

def collapse_record(record, separator = '.', root = None):
    """See :func:`brewery.expand_record` for reverse operation.
    """

    result = {}
    for key, value in record.items():
        if root:
            collapsed_key = root + separator + key
        else:
            collapsed_key = key
        
        if type(value) == dict:
            collapsed = collapse_record(value, separator, collapsed_key)
            result.update(collapsed)
        else:
            result[collapsed_key] = value
    return result
    
class Field(object):
    """Metadata - information about a field in a dataset or in a datastream.

    :Attributes:
        * `name` - field name
        * `label` - optional human readable field label
        * `storage_type` - Normalized data storage type. The data storage type is abstracted
        * `concrete_storage_type` (optional, recommended) - Data store/database dependent storage
            type - this is the real name of data type as used in a database where the field
            comes from or where the field is going to be created (this might be null if unknown)
        * `analytical_type` - data type used in data mining algorithms
        * `missing_values` (optional) - Array of values that represent missing values in the
            dataset for given field
    """

    storage_types = ["unknown", "string", "text", "integer", "float", "boolean", "date"]
    analytical_types = ["default", "typeless", "flag", "discrete", "range", 
                        "set", "ordered_set"]

    default_analytical_type = {
                    "unknown": "typeless",
                    "string": "typeless",
                    "text": "typeless",
                    "integer": "discrete",
                    "float": "range",
                    "date": "typeless"
                }

    def __init__(self, name, label = None, storage_type = "unknown", analytical_type = "typeless", 
                    concrete_storage_type = None, missing_values = None):
        self.name = name
        self.label = label
        self.storage_type = storage_type
        self.analytical_type = analytical_type
        self.concrete_storage_type = concrete_storage_type
        self.missing_values = missing_values

    def __str__(self):
        """Return field name as field string representation."""
        
        return self.name

    def __repr__(self):
        d = {}
        d["name"] = self.name
        d["label"] = self.label
        d["storage_type"] = self.storage_type
        d["analytical_type"] = self.analytical_type
        d["concrete_storage_type"] = self.concrete_storage_type
        d["missing_values"] = self.missing_values
        return "<%s(%s)>" % (self.__class__, d)

    def __eq__(self, other):
        if self is other:
            return True
        if self.name != other.name or self.label != other.label:
            return False
        elif self.storage_type != other.storage_type or self.analytical_type != other.analytical_type:
            return False
        elif self.concrete_storage_type != other.concrete_storage_type:
            return False
        elif self.missing_values != other.missing_values:
            return False
        else:
            return True
            
    def __ne__(self,other):
        return not self.__eq__(other)

class FieldList(object):
    """List of fields"""
    def __init__(self, fields = None):
        """
        Create a list of :class:`Field` objects from a list of strings, dictionaries or tuples

        How fields are consutrcuted:

        * string: `field name` is set 
        * tuple: (`field_name`, `storaget_type`, `analytical_type`), the `field_name` is
          obligatory, rest is optional
        * dict: contains key-value pairs for initializing a :class:`Field` object

        For strings and in if not explicitly specified in a tuple or a dict case, then following rules
        apply:

        * `storage_type` is set to ``unknown``
        * `analytical_type` is set to ``typeless``
        """
        super(FieldList, self).__init__()

        self._fields = []
        self._field_dict = {}
        self._field_names = []

        if fields:
            for field in fields:
                self.append(field)
        
    def append(self, field):
        """Add field to list of fields.
        
        :Parameters:
            * `field` - :class:`Field` object, ``str``, ``tuple`` or ``dict`` object 

        If field is not a `Field` object, then construction of new field is as follows:

        * ``str``: `field name` is set 
        * ``tuple``: (`field_name`, `storaget_type`, `analytical_type`), the `field_name` is
          obligatory, rest is optional
        * ``dict``: contains key-value pairs for initializing a :class:`Field` object

        For strings and in if not explicitly specified in a tuple or a dict case, then following rules
        apply:

        * `storage_type` is set to ``unknown``
        * `analytical_type` is set to ``typeless``
        """


        d = {}
        d["storage_type"] = "unknown"
        d["analytical_type"] = "typeless"

        if type(field) == Field:
            # FIXME: should be a copy?
            new_field = field
        else:
            if type(field) == str or type(field) == unicode:
                d["name"] = field
            elif type(field) == tuple or type(field) == list:
                d["name"] = field[0]
                if len(field) > 1:
                    d["storage_type"] = field[1]
                    if len(field) > 2:
                        d["analytical_type"] = field[2]
            elif type(field) == dict:
                d["name"] = field["name"]
                if "label" in field:
                    d["label"] = field["label"]
                if "storage_type" in field:
                    d["storage_type"] = field["storage_type"]
                if "analytical_type" in field:
                    d["analytical_type"] = field["analytical_type"]
                if "adapter_storage_type" in field:
                    d["adapter_storage_type"] = field["adapter_storage_type"]
            else:
                raise ValueError("Unknown field object type ('%s' ) of field description object '%s'" \
                                    % (type(field), field))

            if "analytical_type" not in d:
                deftype = Field.default_analytical_types[d["storage_type"]]
                d["analytical_type"] = deftype

            new_field = Field(**d)
            
        self._fields.append(new_field)
        self._field_dict[new_field.name] = new_field
        self._field_names.append(new_field.name)
        
    def names(self, indexes = None):
        """Return names of fields in the list.

        :Parameters:
            * `indexes` - list of indexes for which field names should be collected. If set to
              ``None`` then all field names are collected - this is default behaviour.
        """
        
        if indexes:
            names = []
            for i in indexes:
                names.append(self._field_names[i])
            return names
        else:
            return self._field_names

    def indexes(self, fields):
        """Return a tuple with indexes of fields from ``fields`` in a data row. Fields
        should be a list of ``Field`` objects or strings.
        
        This method is useful when it is more desirable to process data as rows (arrays), not as
        dictionaries, for example for performance purposes.
        """

        indexes = [self.index(str(field)) for field in fields]

        return tuple(indexes)

    def index(self, field):
        """Return index of a field"""
        
        try:
            index = self._field_names.index(str(field))
        except ValueError:
            raise KeyError("Field list has no field with name '%s'" % str(field))

        return index

    def fields(self, names = None):
        """Return a tuple with fields. `names` specifies which fields are returned. When names is
        ``None`` all fields are returned.
        """

        if not names:
            return self._fields

        fields = []
        for name in names:
            if name in self._field_dict:
                fields.append(self._field_dict[name])
            else:
                raise KeyError("Field list has no field with name '%s'" % name)

        return fields

    def field(self, name):
        """Return a field with name `name`"""

        if name in self._field_dict:
            return self._field_dict[name]
        raise KeyError("Field list has no field with name '%s'" % name)
    
    def __len__(self):
        return len(self._fields)
        
    def __getitem__(self, index):
        return self._fields[index]
        
    def __setitem__(self, index, new_field):
        field = self._fields[index]
        del self._field_dict[field.name]
        self._fields[index] = new_field
        self._field_names[index] = new_field.name
        self._field_dict[new_field.name] = new_field
        
    def __delitem__(self, index):
        field = self._fields[index]
        del self._field_dict[field.name]
        del self._fields[index]
        del self._field_names[index]
        
    def __iter__(self):
        return self._fields.__iter__()
        
    def __contains__(self, field):
        if type(field) == str or type(field) == unicode:
            return field in self._field_names
            
        return field in self._fields

    def __iconcat__(self, array):
        for field in array:
            self.append(field)

    def __concat__(self, array):
        fields = self.copy()
        fields += array
        return fields
        
    def copy(self, fields = None):
        """Return a shallow copy of the list.
        
        :Parameters:
            * `fields` - list of fields to be copied.
        """
        if fields is not None:
            copy_fields = self.fields(fields)
            return FieldList(copy_fields)
        else:
            return FieldList(self._fields)
            
class FieldMap(object):
    """Filters fields in a stream"""
    def __init__(self, rename = None, drop = None):
        super(FieldMap, self).__init__()
        if rename:
            self.rename = rename
        else:
            self.rename = {}
        if drop:
            self.drop = drop
        else:
            self.drop = []
        
    def map(self, fields):
        """Map `fields` according to the FieldMap: rename or drop fields as specified. Returns
        a FieldList object."""
        output_fields = FieldList()
        
        for field in fields:
            if field.name in self.rename:
                # Create a copy and rename field if it is mapped
                new_field = copy.copy(field)
                new_field.name = self.rename[field.name]
            else:
                new_field = field

            if field.name not in self.drop:
                # Pass field if it is not in dropped field list
                output_fields.append(new_field)
            
        return output_fields


    def row_filter(self, fields):
        """Returns an object that will convert rows with structure specified in `fields`. You can
        use the object to filter fields from a row (list, array) according to this map."""
        indexes = []
        
        for i, field in enumerate(fields):
            if field.name not in self.drop:
                indexes.append(i)
                
        return RowFieldFilter(indexes)
        
class RowFieldFilter(object):
    """Class for filtering fields in array"""

    def __init__(self, indexes = None):
        """Create an instance of RowFieldFilter. `indexes` is a list of indexes that are passed
        to output."""
        super(RowFieldFilter, self).__init__()
        if indexes is not None:
            self.indexes = indexes
        else:
            self.indexes = []
        
    def filter(self, row):
        """Filter a `row` according to ``indexes``."""
        nrow = []
        for i in self.indexes:
            nrow.append(row[i])
        return nrow
        



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Data stores, data sets and data sources
"""

# Data sources
# ============
# 
# Should implement:
# * fields
# * prepare()
# * rows() - returns iterable with value tuples
# * records() - returns iterable with dictionaries of key-value pairs
# 
# Data targets
# ============
# Should implement:
# * fields
# * prepare()
# * append(object) - appends object as row or record depending whether it is a dictionary or a list
# Optional (for performance):
# * append_row(row) - row is tuple of values, raises exception if there are more values than fields
# * append_record(record) - record is a dictionary, raises exception if dict key is not in field list

import urllib2
import urlparse
import brewery.dq
import copy

def open_resource(resource, mode = None):
    """Get file-like handle for a resource. Conversion:
    
    * if resource is a string and it is not URL or it is file:// URL, then opens a file
    * if resource is URL then opens urllib2 handle
    * otherwise assume that resource is a file-like handle
    
    Returns tuple: (handle, should_close) where `handle` is file-like object and `should_close` is
        a flag whether returned handle should be closed or not. Closed should be resources which
        where opened by this method, that is resources referenced by a string or URL.
    """
    handle = None
    should_close = False
    if type(resource) == str or type(resource) == unicode:
        parts = urlparse.urlparse(resource)
        if parts.scheme == '' or parts.scheme == 'file':
            if mode:
                handle = file(resource, mode)
                should_close = True
            else:
                handle = file(resource)
                should_close = True
        else:
            handle = urllib2.urlopen(resource)
            should_close = True
    else:
        handle = resource
    return (handle, should_close)

class DataStream(object):
    """Shared methods for data targets and data sources"""
    
    def initialize(self):
        """Delayed stream initialisation code. Subclasses might override this method to implement
        file or handle opening, connecting to a database, doing web authentication, ... By
        default this method does nothing.
        
        The method does not take any arguments, it expects pre-configured object.
        """
        pass

    def finalize(self):
        """Subclasses might put finalisation code here, for example:
        
        * closing a file stream
        * sending data over network
        * writing a chart image to a file
        
        Default implementation does nothing.
        """
        pass

    def __get_fields(self):
        """Information about fields: tuple of :class:`Field` objects representing fields passed
        through the receiving stream - either read from data source (:meth:`DataSource.rows`) or written
        to data target (:meth:`DataTarget.append`).

        Subclasses should implement `fields` property getter. Implementing `fields` setter is optional.

        Implementation of `fields` setter is recommended for :class:`DataSource` subclasses such as CSV
        files or typeless document based database. For example: explicitly specify field names for CSVs
        without headers or for specify field analytical or storage types for further processing. Setter
        is recommended also for :class:`DataTarget` subclasses that create datasets (new CSV file,
        non-existing tables).
        """
        return self._fields

    def __set_fields(self, fields):
        self._fields = fields
        # raise Exception("Data stream %s does not support setting fields." % str(self.__class__))

    fields = property(__get_fields, __set_fields)

    @property
    def field_names(self):
        """Returns list of field names. This is shourt-cut for extracting field.name attribute from
        list of field objects returned by :meth:`fields`.
        """
        return [field.name for field in self.fields]

class DataSource(DataStream):
    """Input data stream - for reading."""

    def rows(self):
        """Return iterable object with tuples. This is one of two methods for reading from
        data source. Subclasses should implement this method.
        """
        raise NotImplementedError()

    def records(self):
        """Return iterable object with dict objects. This is one of two methods for reading from
        data source. Subclasses should implement this method.
        """
        raise NotImplementedError()

    def read_fields(self, limit = 0, collapse = False):
        """Read field descriptions from data source. You should use this for datasets that do not
        provide metadata directly, such as CSV files, document bases databases or directories with
        structured files. Does nothing in relational databases, as fields are represented by table
        columns and table metadata can obtained from database easily.
        
        Note that this method can be quite costly, as by default all records within dataset are read
        and analysed.
        
        After executing this method, stream ``fields`` is set to the newly read field list and may
        be configured (set more appropriate data types for example).
        
        :Arguments:
            - `limit`: read only specified number of records from dataset to guess field properties
            - `collapse`: whether records are collapsed into flat structure or not
            
        Returns: tuple with Field objects. Order of fields is datastore adapter specific.
        """

        keys = []
        probes = {}

        def probe_record(record, parent = None):
            for key, value in record.items():
                if parent:
                    full_key = parent + "." + key
                else:
                    full_key = key

                if self.expand and type(value) == dict:
                    probe_record(value, full_key)
                    continue

                if not full_key in probes:
                    probe = brewery.dq.FieldTypeProbe(full_key)
                    probes[full_key] = probe
                    keys.append(full_key)
                else:
                    probe = probes[full_key]
                probe.probe(value)

        count = 0
        for record in self.records():
            if collapse:
                record = collapse_record(record)
            print record
            probe_record(record)
            if limit and count >= limit:
                break
            count += 1

        fields = []

        for key in keys:
            probe = probes[key]
            field = Field(probe.field)

            storage_type = probe.unique_storage_type
            if not storage_type:
                field.storage_type = "unknown"
            elif storage_type == "unicode":
                field.storage_type = "string"
            else:
                field.storage_type = "unknown"
                field.concrete_storage_type = storage_type

            # FIXME: Set analytical type

            fields.append(field)

        self._fields = list(fields)
        return self._fields

class DataTarget(DataStream):
    """Output data stream - for writing.
    """

    def append(self, object):
        """Append an object into dataset. Object can be a tuple, array or a dict object. If tuple
        or array is used, then value position should correspond to field position in the field list,
        if dict is used, the keys should be valid field names.        
        """
        raise NotImplementedError()
     
